--[[
   Arms Race
   v1.0

   Some game type Ryoko told me to implement

   by irons (jonirons@gmail.com)
]]

--[[
   customize weapon order and allowed item pickups using
   the tables below
]]

-- the weapons, last to first, used when progressing
-- through the game
OrderedWeapons = {
   
   --[[
      highest in list = last (winning) 
      lowest in list = starting weapon
      each repeated entry = add one to inventory
          (e.g. shotguns are in default list twice)
      consecutive repeated (n) entries = single "level" with
          n in inventory
   ]]
   
   "fist",
   "pistol", "pistol", -- 2x pistol in same level
   "alien weapon",
   "fusion pistol",
   "assault rifle",
   "shotgun",
   "smg",
   "flamethrower",
   "shotgun",
   "missile launcher"
}

-- items that can still spawn.
-- others (by default ammo and invincibility)
-- are left out
KeepTheseItems = {
   "invisibility",
   "infravision",
   "extravision",
   "oxygen",
   "single health",
   "double health",
   "triple health",
   "key",
   "uplink chip",
   "balls" -- a special keyword, shorthand for all ball colors
}

-- items that the player should not always
-- have (generally powerups, chips, balls...)
-- otherwise, player will always have 1
DoNotGiveTheseItems = {
   "knife",
   "invisibility",
   "invincibility",
   "infravision",
   "extravision",
   "oxygen",
   "single health",
   "double health",
   "triple health",
   "key",
   "uplink chip",
   "balls" -- a special keyword, shorthand for all ball colors
}

--[[
   modify below this comment at your own risk




























]]

KeyWeapons = {}
WeaponsTemplate = {}
KeptItems = {}
NonDefaultItems = {}
AllBalls = {
      "ball",
      "violet ball",
      "yellow ball",
      "brown ball",
      "orange ball",
      "blue ball",
      "green ball",
      "light blue ball"
}
ProjectileWeapons = {}
WeaponProjectiles = {}
EndGameCounter = nil
Triggers = {}

function initKeyWeapons()
   for i, itemType in ipairs(OrderedWeapons) do
      KeyWeapons[itemType] = true
   end
end

function initLackOfNonKeptItems()
   local itemTypeString
   for itemType in ItemTypes() do
      itemTypeString = tostring(itemType)
      if not KeptItems[itemTypeString] then
	 itemType.initial_count = 0
	 itemType.maximum_count = 0
	 itemType.minimum_count = 0
	 itemType.random_chance = 0
	 itemType.random_location = false
	 itemType.total_available = 0
      end
   end
end

function initKeptItems()
   for i, item in ipairs(KeepTheseItems) do
      if "balls" == item then
	 for i, ball in ipairs(AllBalls) do
	    KeptItems[ball] = true
	 end
      else
	 KeptItems[item] = true
      end
   end
end

function initNonDefaultItems()
   for i, item in ipairs(DoNotGiveTheseItems) do
      if "balls" == item then
	 for i, ball in ipairs(AllBalls) do
	    NonDefaultItems[ball] = true
	 end
      else
	 NonDefaultItems[item] = true
      end
   end
end

function initPlayer(player)
   
   local previousWeapon
   local quantities = {}
   local newTemplate = #WeaponsTemplate == 0
   player._weapons = {}
   local weaponString
   
   for i, weapon in ipairs(OrderedWeapons) do

      local weaponQuantity = {}
      if not quantities[weapon] then
	 quantities[weapon] = 1
      else
	 quantities[weapon] = quantities[weapon] + 1
      end
      
      weaponQuantity.weapon = weapon
      weaponQuantity.quantity = quantities[weapon]

      -- consecutive = multiple in single level
      if previousWeapon and weapon == previousWeapon then
	 table.remove(player._weapons)
	 quantities[weapon] = 0
      end
      
      table.insert(player._weapons, weaponQuantity)
      previousWeapon = weapon
   end

   player._lastSanctioned = getSanctionedWeapon(player)
   
   if newTemplate then
      for i, weaponQuantity in ipairs(player._weapons) do
	 table.insert(WeaponsTemplate, weaponQuantity)
      end
   end

   for item in ItemTypes() do
      player.items[tostring(item)] = 0
   end
   
end

-- starting outside of init() because we need to prevent spawn

initKeptItems()
initLackOfNonKeptItems()

function Triggers.init()

   Game.over = false
   initKeyWeapons()
   initNonDefaultItems()
   
   for player in Players() do
     initPlayer(player)         
   end
   
end

function Triggers.idle()

   for player in Players() do
      suicideWatch(player)
      constantItems(player)
      forceSanctionedWeapon(player)
      unsanctionedPlayer(player)
   end

   for projectile in Projectiles() do
      trackProjectileWeapons(projectile)
   end

   endGameCounter()
   
end

function trackProjectileWeapons(projectile)
   if projectile and not projectile._tracked then
      if projectile.owner then
	 local player = projectile.owner.player
	 if player and not player.dead then
	    local projectileString = tostring(projectile.type)
	    local weaponType = player.weapons.current.type
	    local weaponString = tostring(weaponType)
	    if weaponType then
	       ProjectileWeapons[projectileString] = weaponString
	    end
	 end
      end
      projectile._tracked = true
   end
end

function forceSanctionedWeapon(player)
   
   local sanctionedWeapon = getSanctionedWeapon(player)
   
   if sanctionedWeapon then

      if not player.dead then

	 player.action_flags.cycle_weapons_backward = false
	 player.action_flags.cycle_weapons_forward = false

	 local sanctionedWeaponType = sanctionedWeapon.weapon

	 local stringVal = tostring(sanctionedWeaponType)

	 if sanctionedWeaponType and player.weapons[stringVal] then
	    player.weapons[stringVal]:select()
	 end
	 
      end
   else
      if not player.dead then
	 player.weapons.active = false
      end
   end
end

function constantItems(player)

   local sanctionedWeapon = getSanctionedWeapon(player)

   if sanctionedWeapon then
      local sanctionedWeaponType = sanctionedWeapon.weapon
      local sanctionedQuantity = sanctionedWeapon.quantity
      
      local sanctionedString = tostring(sanctionedWeaponType)

      if not player.dead then
	 local itemString
	 local numberItems
	 for itemType in ItemTypes() do
	    itemString = tostring(itemType)
	    numberItems = player.items[itemString]
	    if not NonDefaultItems[itemString] then
	       if not KeyWeapons[itemString] and numberItems < 1 then
		  player.items[itemString] = 100
	       else
		  if sanctionedWeapon and sanctionedString == itemString
		     and numberItems ~= sanctionedQuantity
		  then
		     player.items[itemString] = sanctionedQuantity
		  else
		     if not sanctionedWeapon and numberItems ~= 0 then
			player.items[itemString] = 0
		     end
		  end
	       end
	    end
	 end
      end
   else
      if not player.dead then
	 local itemString
	 for itemType in ItemTypes() do
	    itemString = tostring(itemType)
	    numberItems = player.items[itemString]
	    if numberItems ~= 0 then
	       player.items[itemString] = 0
	    end
	 end
      end
   end
end

function playerKilledSelf(player, projectile)
   startSuicideWatch(player, tostring(projectile))
end

function handleSuicide(player, projectileType)
   
   if player._lastSanctioned then
      
      local weaponType = findWeaponType(projectileType, player)
      
      local sanctionedType = tostring(player._lastSanctioned.weapon)
      
      -- if the player advanced because of this suicide...
      if sanctionedType ~= weaponType or player.points == 1 then
	 
	 local currentSanctioned = getSanctionedWeapon(player)

	 if currentSanctioned.weapon ~= player._lastSanctioned.weapon or
	    currentSanctioned.quantity ~= player._lastSanctioned.quantity
	 then
	    table.insert(player._weapons, player._lastSanctioned)
	 end

	 local achievement = getAchievementString(player._lastSanctioned)

	 if not player._weaponKills then
	    player._weaponKills = {}
	 end
	 
	 if player._weaponKills[achievement] then
	    player._weaponKills[achievement] = false
	 end

	 if player.points > 0 then
	    player.points = player.points - 1
	 end
	 
      end
   end
   player._suicideWatch = nil
end

function startSuicideWatch(player, projectileType)
   player._suicideWatch = projectileType
end

function suicideWatch(player)
   if player._suicideWatch then
      handleSuicide(player, player._suicideWatch)
   end
end

function findWeaponType(projectileType, player)
      local weaponType = ProjectileWeapons[projectileType]

      if not weaponType then

	 if not player.dead then
	    weaponType = 
	       tostring(player.weapons.current.type)
	 else
	    weaponType = tostring(player._lastSanctioned.weapon)
	 end
	 
	 local weaponTypeExists

	 for pt, wt in pairs(ProjectileWeapons) do
	    if wt == weaponType then
	       weaponTypeExists = true
	       break
	    end
	 end

	 if not weaponTypeExists then
	    ProjectileWeapons[projectileType] = weaponType
	 end
      end
      return weaponType
end

function playerKilledSomeone(player, projectile)
   if player and projectile then

      local projectileType = tostring(projectile.type)
      local weaponType = findWeaponType(projectileType, player)
   
      if not player._weaponKills then
	 player._weaponKills = {}
      end

      local sanctioned = getSanctionedWeapon(player)
      
      if sanctioned then

	 player._lastSanctioned = sanctioned

	 local sanctionedWeapon = tostring(sanctioned.weapon)	 	 
	 local achievement = getAchievementString(sanctioned)
	 
	 if not player._weaponKills[achievement]
	    and weaponType == sanctionedWeapon
	 then
	    advancePlayer(player)
	    player._weaponKills[achievement] = true
	 end
	 
      end
      
   end   
end

function getAchievementString(weaponQuantity)
   return weaponQuantity.quantity .. " * "
      .. tostring(weaponQuantity.weapon)
end

function Triggers.monster_killed(victim, aggressor, projectile)
   if victim and not victim.player and aggressor then
      local player = aggressor
      if aggressor ~= victim then
	 playerKilledSomeone(player, projectile)
      else
	 playerKilledSelf(player, projectile)
      end
   end
end

function Triggers.player_killed(victim, aggressor, action, projectile)
   if aggressor and victim then
      local player = aggressor
      if aggressor ~= victim and aggressor.team ~= victim.team then
	 playerKilledSomeone(player, projectile)
      else
	 playerKilledSelf(player, projectile)
      end
   end
end

function getSanctionedWeapon(player)
   return player._weapons[#player._weapons]
end

function advancePlayer(player)
   table.remove(player._weapons)
   player.points = player.points + 1
   local sum = #player._weapons
   local teams = false
   for other in Players() do
      if other ~= player then
	 if other.team == player.team then
	    teams = true
	    sum = sum + #other._weapons
	 end
      end
   end
   
   if sum == 0 then
      windGameDown(player, teams)
   end
end

function windGameDown(player, teamGame)
   EndGameCounter = 150
   for dude in Players() do
      dude:play_sound("alarm", 0.5)
   end
   if not teamGame then
      Players.print(player.name .. " won the game!")
   else
      Players.print(player.name .. " won the game for " .. player.color
		       .. " team!")
   end
end

function unsanctionedPlayer(player)
   local sanctioned = #player._weapons > 0
   if not sanctioned then
      player.energy = 450
      player.invincibility_duration = 120
   end
end

function endGameCounter()
   if EndGameCounter then
      if EndGameCounter > 0 then
	 EndGameCounter = EndGameCounter - 1
      else
	 Game.over = true
      end
   end
end
